// MIKE (C) http://mikelab.kiev.ua/

unit Metters;

interface

uses
  Controls, Classes, Graphics;

type
  TAnalogMetter = class(TGraphicControl)
  strict private
    FMin, FMax, FValue: Double;
    FLabelsFormat, FValueFormat, FMetterType, FSubType: string;
    FLabelsCount, FDemphTime: Integer;
    FScaleColor, FBackColor, FLabelColor, FValueColor, FPointerColor: TColor;
    FFont: TFont;
    FUpdating: Boolean;
    FOldFontChange: TNotifyEvent;

    procedure   OnFontChange(Sender: TObject);
    procedure   SetLabelsFormat(const Format: string);
    procedure   SetValueFormat(const Format: string);
    procedure   SetMetterType(const Value: string);
    procedure   SetSubType(const Value: string);
    procedure   SetScaleColor(const Color: TColor);
    procedure   SetBackColor(const Color: TColor);
    procedure   SetLabelColor(const Color: TColor);
    procedure   SetValueColor(const Color: TColor);
    procedure   SetPointerColor(const Color: TColor);
    procedure   SetMin(const Value: Double);
    procedure   SetMax(const Value: Double);
    procedure   SetValue(const Value: Double);
    procedure   SetLabelsCount(const Count: Integer);
  strict protected
    procedure   DrawBack; virtual; abstract;
    procedure   DrawScale; virtual; abstract;
    procedure   DrawValue; virtual; abstract;

    property    LabelsFormat: string read FLabelsFormat write SetLabelsFormat;
    property    ValueFormat: string read FValueFormat write SetValueFormat;
    property    MetterType: string read FMetterType write SetMetterType;
    property    SubType: string read FSubType write SetSubType;
    property    ScaleColor: TColor read FScaleColor write SetScaleColor default clBlack;
    property    BackColor: TColor read FBackColor write SetBackColor default clWhite;
    property    LabelColor: TColor read FLabelColor write SetLabelColor default clBlue;
    property    ValueColor: TColor read FValueColor write SetValueColor default clRed;
    property    PointerColor: TColor read FPointerColor write SetPointerColor default clBlack;
    property    Min: Double read FMin write SetMin;
    property    Max: Double read FMax write SetMax;
    property    Value: Double read FValue write SetValue;
    property    LabelsCount: Integer read FLabelsCount write SetLabelsCount default 10;
    property    Width default 100;
    property    Height default 50;
    property    Font: TFont read FFont;
  public
    constructor Create(AOwner: TComponent); override;
    procedure   SetScaleBounds(const Min, Max: Double);
    procedure   Paint; override;
  end;

  TArcMetter = class(TAnalogMetter)
  strict protected
    procedure   DrawBack; override;
    procedure   DrawScale; override;
    procedure   DrawValue; override;
  published
    property    LabelsFormat;
    property    ValueFormat;
    property    MetterType;
    property    SubType;
    property    ScaleColor;
    property    BackColor;
    property    LabelColor;
    property    ValueColor;
    property    PointerColor;
    property    Min;
    property    Max;
    property    Value;
    property    LabelsCount;
    property    Width;
    property    Height;
    property    Font;
  end;

  TTermMetter = class(TAnalogMetter)
  strict private
    FTubeColor: TColor;

    procedure   SetTubeColor(const Color: TColor);
  strict protected
    procedure   DrawBack; override;
    procedure   DrawScale; override;
    procedure   DrawValue; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property    LabelsFormat;
    property    ValueFormat;
    property    SubType;
    property    ScaleColor;
    property    BackColor;
    property    LabelColor;
    property    PointerColor;
    property    ValueColor;
    property    Min;
    property    Max;
    property    Value;
    property    LabelsCount;
    property    Width;
    property    Height;
    property    Font;
    
    property    TubeColor: TColor read FTubeColor write SetTubeColor default clSilver;
  end;

  TRezMetter = class(TAnalogMetter)
  strict private
    FBarColor: TColor;

    procedure   SetBarColor(const Color: TColor);
  strict protected
    procedure   DrawBack; override;
    procedure   DrawScale; override;
    procedure   DrawValue; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property    LabelsFormat;
    property    ValueFormat;
    property    MetterType;
    property    SubType;
    property    ScaleColor;
    property    BackColor;
    property    LabelColor;
    property    ValueColor;
    property    Min;
    property    Max;
    property    Value;
    property    LabelsCount;
    property    Width;
    property    Height;
    property    Font;

    property    BarColor: TColor read FBarColor write SetBarColor default clSilver;
  end;

implementation

uses
  SysUtils, Windows;

{ TAnalogMetter }

constructor TAnalogMetter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOldFontChange:= Canvas.Font.OnChange;
  FFont:= Canvas.Font;
  FFont.Name:= 'Arial';
  FFont.Size:= 6;
  FFont.OnChange:= OnFontChange;
  Width:= 100;
  Height:= 50;
  FMin:= 0;
  FMax:= 100;
  FValue:= 0;
  FLabelsFormat:= '0';
  FValueFormat:= '0.0';
  FMetterType:= 'M';
  FSubType:= 'x1';
  FLabelsCount:= 10;
  FScaleColor:= clBlack;
  FBackColor:= clWhite;
  FLabelColor:= clBlue;
  FValueColor:= clRed;
  FPointerColor:= clBlack;
  FUpdating:= False;
  FDemphTime:= 0;
end;

procedure TAnalogMetter.OnFontChange(Sender: TObject);
begin
  FOldFontChange(Self);
  if not FUpdating then Repaint;
end;

procedure TAnalogMetter.Paint;
var
  OldFontColor: TColor;
begin
  if Assigned(Parent) then
    begin
      FUpdating:= True;
      try
        OldFontColor:= FFont.Color;
        inherited Paint;
        DrawBack;
        DrawScale;
        DrawValue;
        FFont.Color:= OldFontColor;
      finally
        FUpdating:= False;
      end;
    end;
end;

procedure TAnalogMetter.SetBackColor(const Color: TColor);
begin
  if Color <> FBackColor then
    begin
      FBackColor:= Color;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetLabelColor(const Color: TColor);
begin
  if Color <> FLabelColor then
    begin
      FLabelColor:= Color;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetLabelsCount(const Count: Integer);
begin
  if (Count <> FLabelsCount) and (Count > 0) then
    begin
      FLabelsCount:= Count;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetLabelsFormat(const Format: string);
begin
  if Format <> FLabelsFormat then
    begin
      FLabelsFormat:= Format;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetMax(const Value: Double);
begin
  if (Value > FMin) and (Value <> FMax) then
    begin
      FMax:= Value;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetMetterType(const Value: string);
begin
  if Value <> FMetterType then
    begin
      FMetterType:= Value;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetMin(const Value: Double);
begin
  if (Value < FMax) and (Value <> FMin) then
    begin
      FMin:= Value;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetPointerColor(const Color: TColor);
begin
  if Color <> FPointerColor then
    begin
      FPointerColor:= Color;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetValue(const Value: Double);
var
  i, Step: Double;
  Pause: Integer;
begin
  if Value <> FValue then
    begin
      if FDemphTime > 0 then
        begin
          Pause:= Trunc(FDemphTime/(FLabelsCount*Abs((FMax-FMin)/(Value-FValue))));
          Step:= (FMax-FMin)/FLabelsCount;
          i:= FValue+Step;
          while i < Value do
            begin
              FValue:= i;
              Repaint;
              i:= i+Step;
              Sleep(Pause);
            end;
        end;
      FValue:= Value;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetScaleBounds(const Min, Max: Double);
begin
  if (Min < Max) and ((Min <> FMin) or (Max <> FMax)) then
    begin
      FMin:= Min;
      FMax:= Max;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetScaleColor(const Color: TColor);
begin
  if Color <> FScaleColor then
    begin
      FScaleColor:= Color;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetSubType(const Value: string);
begin
  if Value <> FSubType then
    begin
      FSubType:= Value;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetValueColor(const Color: TColor);
begin
  if Color <> FValueColor then
    begin
      FValueColor:= Color;
      Repaint;
    end;
end;

procedure TAnalogMetter.SetValueFormat(const Format: string);
begin
  if Format <> FValueFormat then
    begin
      FValueFormat:= Format;
      Repaint;
    end;
end;

//////////////////////////// TArcMetter /////////////////////////////

procedure TArcMetter.DrawBack;
begin
  if Enabled then
    Canvas.Brush.Color:= BackColor
  else
    Canvas.Brush.Color:= clInactiveCaption;
  Canvas.Brush.Style:= bsSolid;
  Canvas.FillRect(ClientRect);
end;

procedure TArcMetter.DrawScale;
const
  BeginAngle = Pi/8+Pi/2;
  EndAngle = Pi-Pi/8+Pi/2;
var
  Center: TPoint;
  i, Step, Val, ValStep: Double;
  MultX, MultY, x, y, Border: Integer;
  s: string;
begin
  Canvas.Brush.Style:= bsClear;
  Canvas.Pen.Style:= psSolid;
  Canvas.Pen.Width:= 2;
  Canvas.Pen.Color:= ScaleColor;

  Center.X:= Width div 2;
  Center.Y:= Height;
  MultX:= Width-(Width div 2);
  MultY:= Height-(Height div 10);
  if Width >= Height then
    Border:= Height div 20
  else
    Border:= Width div 20;

  i:= BeginAngle;
  repeat
    Canvas.MoveTo(Trunc(Center.X+Sin(i)*MultX), Trunc(Center.Y+Cos(i)*MultY));
    i:= i+3/MultX;
    Canvas.LineTo(Trunc((Center.X+Sin(i)*MultX)), Trunc((Center.Y+Cos(i)*MultY)));
  until i >= EndAngle;

  Canvas.Font.Color:= LabelColor;
  Canvas.Pen.Width:= 1;
  if Enabled then
    begin
      Step:= (EndAngle-BeginAngle)/LabelsCount;
      ValStep:= (Max-Min)/LabelsCount;
      Val:= Max;
      i:= BeginAngle;
      repeat
        s:= FormatFloat(LabelsFormat, Val);
        Canvas.MoveTo(Trunc((Center.X+Sin(i)*MultX)), Trunc((Center.Y+Cos(i)*MultY)));
        x:= Trunc((Center.X+Sin(i)*(MultX*0.9)));
        y:= Trunc((Center.Y-(Border*0.1)+Cos(i)*(MultY*0.9)));
        Canvas.LineTo(x, y);
        Canvas.TextOut(x-(Canvas.TextWidth(s) div 2), y, s);
        Val:= Val-ValStep;
        i:= i+Step;
      until i > EndAngle;
    end;

  Canvas.Font.Color:= ScaleColor;
  Canvas.TextOut(Border, Border, MetterType);
  Canvas.TextOut(Width-Canvas.TextWidth(SubType)-Border, Border, SubType);
end;

procedure TArcMetter.DrawValue;
const
  BeginAngle = pi/8+pi/2;
  EndAngle = pi-pi/8+pi/2;
var
  Center: TPoint;
  i, Pos: Double;
  MultX, MultY, x, y, Border: Integer;
  s: string;
begin
  if Enabled then
    begin
      Canvas.Brush.Style:= bsClear;
      Canvas.Pen.Style:= psSolid;
      Canvas.Pen.Width:= 1;
      Canvas.Pen.Color:= PointerColor;
      Center.X:= Width div 2;
      Center.Y:= Height;
      MultX:= Width-(Width div 2);
      MultY:= Height-(Height div 10);
      if Width >= Height then
        Border:= Height div 20
      else
        Border:= Width div 20;
      Pos:= Value;
      if Pos > Max then Pos:= Max;
      if Pos < Min then Pos:= Min;

      s:= FormatFloat(ValueFormat, Value);
      Canvas.Font.Color:= ValueColor;
      Canvas.TextOut(Trunc((Width/2)-(Canvas.TextWidth(s)/2)), Trunc(Height*0.85-Canvas.TextHeight(s)), s);
      i:= EndAngle-(EndAngle-BeginAngle)/(Max-Min)*(Pos-Min);
      x:= Trunc((Center.X+Sin(i)*(MultX*1.05)));
      y:= Trunc((Center.Y+Cos(i)*(MultY*1.05)));
      Canvas.Pen.Color:= PointerColor;
      Canvas.MoveTo(Center.X, Center.Y-Border);
      Canvas.LineTo(x, y);
      Canvas.MoveTo(Center.X+1, Center.Y-Border);
      Canvas.LineTo(x, y);
      Canvas.MoveTo(Center.X-1, Center.Y-Border);
      Canvas.LineTo(x, y);
    end;
end;

//////////////////////////// TTermMetter ///////////////////////////

constructor TTermMetter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SubType:= 'C';
  Width:= 50;
  Height:= 200;
  FTubeColor:= clSilver;
  PointerColor:= clRed;
end;

procedure TTermMetter.DrawBack;
begin
  if Enabled then
    Canvas.Brush.Color:= BackColor
  else
    Canvas.Brush.Color:= clInactiveCaption;
  Canvas.Brush.Style:= bsSolid;
  Canvas.FillRect(ClientRect);
end;

procedure TTermMetter.DrawScale;
var
  ValStep, Val: Double;
  i, ScaleW, ScaleH, Border, VertBorder, Tank, Step, TubeH: Integer;
  s: string;
begin
  Canvas.Pen.Style:= psSolid;
  Canvas.Pen.Width:= 1;
  Canvas.Pen.Color:= FTubeColor;
  Canvas.Brush.Style:= bsSolid;
  Canvas.Brush.Color:= FTubeColor;

  Border:= Width div 20;
  TubeH:= Width div 10;
  VertBorder:= Height div 10;
  Tank:= Width div 3;
  ScaleW:= Trunc((Width*0.8-TubeH-Border)/2);
  ScaleH:= Height-VertBorder-Border-Tank;
  Step:= ScaleH div LabelsCount;

  Canvas.FillRect(Rect(Trunc(Width/2-TubeH/2), VertBorder-Border, Trunc(Width/2+TubeH/2), Height-Tank));

  if Enabled then
    Canvas.Brush.Color:= PointerColor
  else
    Canvas.Brush.Color:= clGray;
  Canvas.Ellipse(Trunc(Width/2-Tank/2), trunc(Height-Border-Tank), Trunc(Width/2+Tank/2), Trunc(Height-Border));

  Canvas.Pen.Color:= ScaleColor;
  Canvas.Brush.Color:= BackColor;
  Canvas.Brush.Style:= bsClear;
  Canvas.Font.Color:= LabelColor;
  if Enabled then
    begin
      ValStep:= (Max-Min)/LabelsCount;
      Val:= Max;
      i:= VertBorder;
      while i <= ScaleH+VertBorder do
        begin
          s:= FormatFloat(LabelsFormat, Val);
          Canvas.MoveTo(Border, i);
          Canvas.LineTo(Border+ScaleW, i);
          Canvas.TextOut(Border+ScaleW-Canvas.TextWidth(s), i-Canvas.TextHeight(s), s);
          Val:= Val-ValStep;
          Inc(i, Step);
        end;
      end;
      
  Canvas.Font.Color:= ScaleColor;
  Canvas.TextOut(Width-Canvas.TextWidth(SubType)-Border, Border, SubType);
end;

procedure TTermMetter.DrawValue;
var
  LiqH, Tank, LiqTop, Border, VertBorder, ScaleH: Integer;
  s: string;
  Pos: Double;
begin
  if Enabled then
    begin
      Canvas.Pen.Style:= psSolid;
      Canvas.Pen.Width:= 1;
      Canvas.Pen.Color:= PointerColor;
      Canvas.Brush.Style:= bsSolid;
      Canvas.Brush.Color:= PointerColor;
      Canvas.Font.Color:= ValueColor;

      Border:= Width div 20;
      VertBorder:= Height div 10;
      LiqH:= (Width div 10)-2;
      Tank:= Width div 3;
      ScaleH:= Height-VertBorder-Border-Tank;
      Pos:= Value;
      if Pos > Max then Pos:= Max;
      if Pos < Min then Pos:= Min;
      LiqTop:= Trunc(Height-Border-Tank-ScaleH/(Max-Min)*(Pos-Min));

      Canvas.FillRect(Rect(Trunc(Width/2-LiqH/2), LiqTop, Trunc(Width/2+LiqH/2), Height-Border-Tank div 2));

      Canvas.Brush.Color:= BackColor;
      Canvas.Brush.Style:= bsClear;
      s:= FormatFloat(ValueFormat, Value);
      Canvas.TextOut(Trunc(Width/2+Width/20+Border), LiqTop-(Canvas.TextHeight(s) div 2), s);
    end;
end;

procedure TTermMetter.SetTubeColor(const Color: TColor);
begin
  if Color <> FTubeColor then
    begin
      FTubeColor:= Color;
      Repaint;
    end;
end;

////////////////////////// TRezMetter ////////////////////////////////////////

constructor TRezMetter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  MetterType:= 'Hz';
  Width:= 150;
  Height:= 50;
  SetScaleBounds(45, 55);
  Value:= 50;
  FBarColor:= clSilver;
end;

procedure TRezMetter.DrawBack;
begin
  if Enabled then
    Canvas.Brush.Color:= BackColor
  else
    Canvas.Brush.Color:= clInactiveCaption;
  Canvas.Brush.Style:= bsSolid;
  Canvas.FillRect(ClientRect);
end;

procedure TRezMetter.DrawScale;
var
  i, Step, BorderH, BorderW, ScaleW: Integer;
  ValStep, Val: Double;
  s: string;
begin
  Canvas.Pen.Style:= psSolid;
  Canvas.Pen.Width:= 1;
  Canvas.Pen.Color:= FBarColor;
  Canvas.Brush.Style:= bsSolid;
  Canvas.Brush.Color:= FBarColor;

  BorderW:= Width div 20;
  BorderH:= Height div 10;
  Step:= Trunc((Width-BorderW*2)/LabelsCount);
  ScaleW:= Trunc(Width-BorderW*2);

  Canvas.FillRect(Rect(BorderW div 2, Trunc(Height*0.4), BorderW+ScaleW+BorderW div 2, Trunc(Height*0.7)));

  Canvas.Pen.Color:= ScaleColor;
  Canvas.Brush.Color:= BackColor;
  Canvas.Brush.Style:= bsClear;
  Canvas.Font.Color:= LabelColor;
  if Enabled then
    begin
      ValStep:= (Max-Min)/LabelsCount;
      Val:= Min;
      i:= BorderW;
      while i <= ScaleW+BorderW do
        begin
          s:= FormatFloat(LabelsFormat, Val);
          Canvas.MoveTo(i, Trunc(Height*0.7));
          Canvas.LineTo(i, Trunc(Height*0.7)+BorderH);
          Canvas.TextOut(i-(Canvas.TextWidth(s) div 2), Trunc(Height*0.7)+BorderH, s);
          Val:= Val+ValStep;
          Inc(i, Step);
        end;
    end;

  Canvas.Font.Color:= ScaleColor;
  Canvas.TextOut(BorderW, BorderH, MetterType);
  Canvas.TextOut(Width-Canvas.TextWidth(SubType)-BorderW, BorderH, SubType);
end;

procedure TRezMetter.DrawValue;
var
  s: string;
  ScaleW, ScaleH, BorderH, BorderW, BarH, MaxBarH, BarCnt: Integer;
  i, BarW, y, x, Step: Double;
begin
  if Enabled then
    begin
      Canvas.Pen.Style:= psSolid;
      Canvas.Pen.Width:= 1;
      Canvas.Pen.Color:= ScaleColor;
      Canvas.Brush.Style:= bsSolid;
      Canvas.Brush.Color:= ScaleColor;
      Canvas.Font.Color:= ValueColor;

      if Odd(Trunc(Max-Min)) then
        BarCnt:= 21
      else
        BarCnt:= 20;
      BorderH:= Height div 10;
      BorderW:= Width div 20;
      ScaleW:= Width-BorderW*2;
      ScaleH:= Trunc(Height*0.3);
      BarW:= ScaleW/BarCnt;
      MaxBarH:= Trunc(ScaleH*0.8);
      Step:= (Max-Min)/BarCnt;

      x:= BorderW/2;
      y:= (Height*0.4)+(ScaleH/2);
      i:= 0;
      while i <= Max-Min do
        begin
          if Abs(Min+i-Value) < Step/2 then
            BarH:= MaxBarH
          else
            if Abs(Min+i-Value) < Step*1.5 then
              BarH:= MaxBarH div 2
            else
              if Abs(Min+i-Value) < Step*2.5 then
                BarH:= MaxBarH div 4
              else
                BarH:= 1;

          if BarH < 1 then BarH:= 1;
          Canvas.FillRect(Rect(Trunc(x), Trunc(y-BarH/2), Trunc(x+BarW-1), Trunc(y+BarH/2)));
          i:= i+Step;
          x:= x+BarW;
        end;

      Canvas.Brush.Color:= BackColor;
      Canvas.Brush.Style:= bsClear;
      s:= FormatFloat(ValueFormat, Value);
      Canvas.TextOut(Trunc(Width/2-Canvas.TextWidth(s)/2), BorderH, s);
    end;
end;

procedure TRezMetter.SetBarColor(const Color: TColor);
begin
  if Color <> FBarColor then
    begin
      FBarColor:= Color;
      Repaint;
    end;
end;

end.

